# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

# I think that the necesary (import from odl.core.util import is_numeric_dtype)should be separated from the doctest import of the __main__  (from odl.core.util.testutils import run_doctests)
# pylint: disable=ungrouped-imports
# Not relevant as although PyTorchTensor inherits from Tensor, it calls the LinearSpace Element in __init__
# pylint: disable=non-parent-init-called
# pylint: disable=super-init-not-called
# pytorch_array_backend triggers this error but numpy_array_backend doesn't?
# pylint: disable=invalid-name

"""PyTorch implementation of tensor spaces."""

# Only for module availability checking
import importlib.util
from os import path
import sys
from sys import argv

import numpy as np

from odl.core.set.space import LinearSpaceElement
from odl.core.space.base_tensors import Tensor, TensorSpace
from odl.core.util import is_numeric_dtype
from odl.core.array_API_support import ArrayBackend, lookup_array_backend

torch_module = importlib.util.find_spec("torch")
if torch_module is not None:
    import torch
    import array_api_compat.torch as xp

    PYTORCH_AVAILABLE = True
else:
    if path.basename(argv[0]) == 'pytest':
        # If running the doctest suite, we should be able to load this
        # module (without running anything) even if Torch is not installed.
        PYTORCH_AVAILABLE = False
        import pytest

        pytest.skip(allow_module_level=True)
    else:
        raise ImportError("You are trying to use the PyTorch backend, but"
                        + " the `torch` dependency is not available."
                        + "\nEither use a different backend, or install"
                        + " a suitable version of Torch." )

__all__ = (
    'PYTORCH_AVAILABLE',
    'PyTorchTensorSpace',
    'pytorch_array_backend'
    
    )
if PYTORCH_AVAILABLE:
    device_strings = ['cpu'] + [f'cuda:{i}' for i in range(torch.cuda.device_count())]


def to_numpy(x : int | float | bool | complex | Tensor | torch.Tensor):
    """Convenience function to convert an element of a Pytorch Tensor Space,
    a python number or a Torch Tensor to a numpy-compatible array/element.

    Args:
        x (scalar or torch.Tensor): input to convert

    Returns:
        scalar or np.ndarray: the converted form of the input, copied to CPU if necessary.
    """
    if isinstance(x, (int, float, bool, complex)):
        return x    
    elif isinstance(x, Tensor):
        return x.data.detach().cpu().numpy()
    elif isinstance(x, torch.Tensor):
        return x.detach().cpu().numpy()
    else:
        raise TypeError(f"The input {x} of type {type(x)} is not supported."
                      + " Please provide a Python Number, an Element of an"
                      + " ODL Pytorch Tensor Space or a torch.Tensor.")

def from_dlpack(x, device='cpu', copy=None):
    """This should theoretically be a stand-in for `from_dlpack` in the Torch instantiation
    of the Array API. That function varies however in behaviour between current PyTorch versions,
    causing numerous failures. So instead, for now we manually implement conversions from the
    alternative backends relevant to ODL (at the moment, NumPy and PyTorch itself).
    """
    if isinstance(x, torch.Tensor):
        if x.device == device and copy != True:
            return x
        # The redundant-looking copy logic is because Torch does not recognize `copy=None`,
        # aka `AVOID_UNNECESSARY_COPY`.
        return x.to(device, copy = True if copy else False)
    elif isinstance(x, np.ndarray):
        return torch.tensor(x, device=torch.device(device))
    else:
        raise NotImplementedError(
            f"With PyTorch {torch.__version__}, currently no way to handle input of type {type(x)}."
        )


if PYTORCH_AVAILABLE:
    try:
        pytorch_array_backend = ArrayBackend(
            impl='pytorch',
            available_dtypes={
                'bool': xp.bool,
                'int8': xp.int8,
                'int16': xp.int16,
                'int32': xp.int32,
                'int64': xp.int64,
                'uint8': xp.uint8,
                'uint16': xp.uint16,
                'uint32': xp.uint32,
                'uint64': xp.uint64,
                'float32': xp.float32,
                'float64': xp.float64,
                'complex64': xp.complex64,
                'complex128': xp.complex128,
            },
            array_namespace=xp,
            array_constructor=xp.asarray,
            from_dlpack=from_dlpack,
            array_type=xp.Tensor,
            make_contiguous=lambda x: x if x.data.is_contiguous() else x.contiguous(),
            identifier_of_dtype=lambda dt: (
                (dt) if dt in [int, bool, float, complex] else str(dt).split('.', maxsplit=1)[-1]
            ),
            available_devices=device_strings,
            to_cpu=lambda x: (
                x if isinstance(x, (int, float, bool, complex)) else x.detach().cpu()
            ),
            to_numpy=to_numpy,
            to_device=lambda x, device: x.to(device),
        )
    except KeyError:
        # PyTest runs modules twice, causing a "duplicate" registration of the backend.
        # Otherwise this should not happen.
        assert "pytest" in sys.modules
        pytorch_array_backend = lookup_array_backend('pytorch')
else:
    pytorch_array_backend = None


class PyTorchTensorSpace(TensorSpace):
    """Set of tensors of arbitrary data type, implemented with PyTorch.

    A tensor is, in the most general sense, a multi-dimensional array
    that allows operations per entry (keep the rank constant),
    reductions / contractions (reduce the rank) and broadcasting
    (raises the rank).
    For non-numeric data type like ``object``, the range of valid
    operations is rather limited since such a set of tensors does not
    define a vector space.
    Any numeric data type, on the other hand, is considered valid for
    a tensor space, although certain operations - like division with
    integer dtype - are not guaranteed to yield reasonable results.

    Under these restrictions, all basic vector space operations are
    supported by this class, along with reductions based on arithmetic
    or comparison, and element-wise mathematical functions.

    This class is implemented using `torch.Tensor`'s as back-end.

    See the `Wikipedia article on tensors`_ for further details.
    See also [Hac2012] "Part I Algebraic Tensors" for a rigorous
    treatment of tensors with a definition close to this one.

    Note also that this notion of tensors is the same as in popular
    Deep Learning frameworks.

    References
    ----------
    [Hac2012] Hackbusch, W. *Tensor Spaces and Numerical Tensor Calculus*.
    Springer, 2012.

    .. _Wikipedia article on tensors: https://en.wikipedia.org/wiki/Tensor
    """

    def __init__(self, shape, dtype='float64', device = 'cpu', requires_grad=False, **kwargs):
        r"""Initialize a new instance.

        Parameters
        ----------
        shape : positive int or sequence of positive ints
            Number of entries per axis for elements in this space. A
            single integer results in a space with rank 1, i.e., 1 axis.
        dtype (str): optional
            Data type of each element. Defaults to 'float64'
        device (str):
            Device on which the data is. Defaults to 'cpu'
        requires_grad (bool):
            Is True if gradients need to be computed for this Tensor using PyTorch's autograd engine, False otherwise.

        Other Parameters
        ----------------
        weighting : optional
            Use weighted inner product, norm, and dist. The following
            types are supported as ``weighting``:

            ``None``: no weighting, i.e. weighting with ``1.0`` (default).

            `Weighting`: Use this weighting as-is. Compatibility
            with this space's elements is not checked during init.

            ``float``: Weighting by a constant.

            array-like: Pointwise weighting by an array.

            This option cannot be combined with ``dist``,
            ``norm`` or ``inner``. It also cannot be used in case of
            non-numeric ``dtype``.

        dist : callable, optional
            Distance function defining a metric on the space.
            It must accept two `PyTorchTensor` arguments and return
            a non-negative real number. See ``Notes`` for
            mathematical requirements.

            By default, ``dist(x, y)`` is calculated as ``norm(x - y)``.

            This option cannot be combined with ``weight``,
            ``norm`` or ``inner``. It also cannot be used in case of
            non-numeric ``dtype``.

        norm : callable, optional
            The norm implementation. It must accept a
            `PyTorchTensor` argument, return a non-negative real number.
            See ``Notes`` for mathematical requirements.

            By default, ``norm(x)`` is calculated as ``inner(x, x)``.

            This option cannot be combined with ``weight``,
            ``dist`` or ``inner``. It also cannot be used in case of
            non-numeric ``dtype``.

        inner : callable, optional
            The inner product implementation. It must accept two
            `PyTorchTensor` arguments and return an element of the field
            of the space (usually real or complex number).
            See ``Notes`` for mathematical requirements.

            This option cannot be combined with ``weight``,
            ``dist`` or ``norm``. It also cannot be used in case of
            non-numeric ``dtype``.

        exponent : positive float, optional
            Exponent of the norm. For values other than 2.0, no
            inner product is defined.

            This option has no impact if either ``dist``, ``norm`` or
            ``inner`` is given, or if ``dtype`` is non-numeric.

            Default: 2.0

        kwargs :
            Further keyword arguments are passed to the weighting
            classes.

        See Also
        --------
        odl.core.space.space_utils.rn : constructor for real tensor spaces
        odl.core.space.space_utils.cn : constructor for complex tensor spaces
        odl.core.space.space_utils.tensor_space :
            constructor for tensor spaces of arbitrary scalar data type

        Notes
        -----
        - A distance function or metric on a space :math:`\mathcal{X}`
          is a mapping
          :math:`d:\mathcal{X} \times \mathcal{X} \to \mathbb{R}`
          satisfying the following conditions for all space elements
          :math:`x, y, z`:

          * :math:`d(x, y) \geq 0`,
          * :math:`d(x, y) = 0 \Leftrightarrow x = y`,
          * :math:`d(x, y) = d(y, x)`,
          * :math:`d(x, y) \leq d(x, z) + d(z, y)`.

        - A norm on a space :math:`\mathcal{X}` is a mapping
          :math:`\| \cdot \|:\mathcal{X} \to \mathbb{R}`
          satisfying the following conditions for all
          space elements :math:`x, y`: and scalars :math:`s`:

          * :math:`\| x\| \geq 0`,
          * :math:`\| x\| = 0 \Leftrightarrow x = 0`,
          * :math:`\| sx\| = |s| \cdot \| x \|`,
          * :math:`\| x+y\| \leq \| x\| +
            \| y\|`.

        - An inner product on a space :math:`\mathcal{X}` over a field
          :math:`\mathbb{F} = \mathbb{R}` or :math:`\mathbb{C}` is a
          mapping
          :math:`\langle\cdot, \cdot\rangle: \mathcal{X} \times
          \mathcal{X} \to \mathbb{F}`
          satisfying the following conditions for all
          space elements :math:`x, y, z`: and scalars :math:`s`:

          * :math:`\langle x, y\rangle =
            \overline{\langle y, x\rangle}`,
          * :math:`\langle sx + y, z\rangle = s \langle x, z\rangle +
            \langle y, z\rangle`,
          * :math:`\langle x, x\rangle = 0 \Leftrightarrow x = 0`.

        Examples
        --------
        Explicit initialization with the class constructor:

        >>> space = PyTorchTensorSpace(3, dtype=float)
        >>> space
        rn(3, 'float64', 'pytorch')
        >>> space.shape
        (3,)
        >>> space.dtype
        torch.float64
        """
        super().__init__(shape, dtype, device, **kwargs)

    ########## Attributes ##########
    @property
    def array_backend(self) -> ArrayBackend:
        return pytorch_array_backend

    @property
    def array_namespace(self):
        """Name of the array_namespace"""
        return xp

    @property
    def element_type(self):
        """Type of elements in this space: `PyTorchTensor`."""
        return PyTorchTensor

    @property
    def impl(self):
        """Name of the implementation back-end: ``'pytorch'``."""
        return 'pytorch'

    ######### public methods #########
    def broadcast_to(self, inp: int|float|complex|list|tuple|torch.Tensor) -> torch.Tensor:
        """Broadcasts the input to the shape of the TensorSpace.
        Also performs a device conversion if necessary

        Args:
            inp (int | float | complex | list | tuple | torch.Tensor): Input to be broadcasted

        Returns:
            torch.Tensor: Returned array broadcasted to the TensorSpace shape and device
        """
        arr = self.array_namespace.broadcast_to(
                    self.array_namespace.asarray(inp, device=self.device),
                    self.shape
                    )
        return arr

    ######### private methods #########


class PyTorchTensor(Tensor):
    """Representation of a `PyTorchTensorSpace` element.

    This is an internal ODL class; it should not directly be instantiated from user code.
    Instead, always use the `.element` method of the `space` for generating these objects."""
    
    def __init__(self, space, data):
        """Initialize a new instance."""
        LinearSpaceElement.__init__(self, space)
        assert(isinstance(data, xp.Tensor)), f"{type(data)=}, should be torch.Tensor"
        if data.dtype != space.dtype:
            data = data.to(space.dtype)
        if data.device != space.device:
            data = data.to(space.device)
        self.__data = data

    @property
    def data(self):
        """The `torch.Tensor` representing the data of ``self``."""
        return self.__data

    @data.setter
    def data(self, value):
        self.__data = value

    def _assign(self, other, avoid_deep_copy):
        """Assign the values of ``other``, which is assumed to be in the
        same space, to ``self``."""
        if avoid_deep_copy:
            self.__data = other.__data
        else:
            self.__data[:] = other.__data

    ######### Public methods #########
    def copy(self):
        """Return an identical (deep) copy of this tensor.

        Parameters
        ----------
        None

        Returns
        -------
        copy : `PyTorchTensor`
            The deep copy

        Examples
        --------
        >>> space = odl.rn(3, impl='pytorch')
        >>> x = space.element([1, 2, 3])
        >>> y = x.copy()
        >>> y == x
        True
        >>> y is x
        False
        """
        return self.space.element(self.data.clone())

    def __getitem__(self, indices):
        """Return ``self[indices]``.

        Parameters
        ----------
        indices : index expression
            Integer, slice or sequence of these, defining the positions
            of the data array which should be accessed.

        Returns
        -------
        values : `PyTorchTensorSpace.dtype` or `PyTorchTensor`
            The value(s) at the given indices. Note that the returned
            object is a writable view into the original tensor, except
            for the case when ``indices`` is a list.

        Examples
        --------
        For one-dimensional spaces, indexing is as in linear arrays:

        >>> space = odl.rn(3, impl='pytorch')
        >>> x = space.element([1, 2, 3])
        >>> x[0]
        1.0
        >>> x[1:]
        rn(2, 'float64', 'pytorch').element([ 2.,  3.])

        In higher dimensions, the i-th index expression accesses the
        i-th axis:

        >>> space = odl.rn((2, 3), impl='pytorch')
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> x[0, 1]
        2.0
        >>> x[:, 1:]
        rn((2, 2), 'float64', 'pytorch').element(
            [[ 2.,  3.],
             [ 5.,  6.]]
        )

        Slices can be assigned to, except if lists are used for indexing:

        >>> y = x[:, ::2]  # view into x
        >>> y[:] = -9
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[-9.,  2., -9.],
             [-9.,  5., -9.]]
        )
        >>> y = x[[0, 1], [1, 2]]  # not a view, won't modify x
        >>> y
        rn(2, 'float64', 'pytorch').element([ 2., -9.])
        >>> y[:] = 0
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[-9.,  2., -9.],
             [-9.,  5., -9.]]
        )
        """
        # Lazy implementation: index the array and deal with it
        if isinstance(indices, type(self)):
            indices = indices.data
        arr = self.data[indices]

        if arr.ndim == 0:
            if self.space.field is not None:
                return self.space.field.element(arr)
            else:
                return arr
        else:
            if is_numeric_dtype(self.dtype):
                weighting = self.space.weighting
            else:
                weighting = None
            space = type(self.space)(
                arr.shape, dtype=self.dtype, exponent=self.space.exponent,
                weighting=weighting, device=self.device)
            return space.element(arr, copy=False)

    def __setitem__(self, indices, values):
        """Implement ``self[indices] = values``.

        Parameters
        ----------
        indices : index expression
            Integer, slice or sequence of these, defining the positions
            of the data array which should be written to.
        values : scalar, array-like or `PyTorchTensor`
            The value(s) that are to be assigned.

            If ``index`` is an integer, ``value`` must be a scalar.

            If ``index`` is a slice or a sequence of slices, ``value``
            must be broadcastable to the shape of the slice.

        Examples
        --------
        For 1d spaces, entries can be set with scalars or sequences of
        correct shape:

        >>> space = PyTorchTensorSpace(3, dtype=float)
        >>> x = space.element([1, 2, 3])
        >>> x[0] = -1
        >>> x[1:] = (0, 1)
        >>> x
        rn(3, 'float64', 'pytorch').element([-1.,  0.,  1.])

        It is also possible to use tensors of other spaces for
        casting and assignment:

        >>> space = PyTorchTensorSpace((2,3), dtype=float)
        >>> x = space.element([[1, 2, 3],
        ...                    [4, 5, 6]])
        >>> x[0, 1] = -1
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[ 1., -1.,  3.],
             [ 4.,  5.,  6.]]
        )
        >>> short_space = PyTorchTensorSpace((2, 2), dtype='int32')
        >>> y = short_space.element([[-1, 2],
        ...                          [0, 0]])
        >>> x[:, :2] = y
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[-1.,  2.,  3.],
             [ 0.,  0.,  6.]]
        )

        The PyTorch assignment and broadcasting rules apply:

        >>> x[:] = torch.tensor([[0, 0, 0],
        ...                  [1, 1, 1]])
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[ 0.,  0.,  0.],
             [ 1.,  1.,  1.]]
        )
        >>> x[:, 1:] = [7, 8]
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[ 0.,  7.,  8.],
             [ 1.,  7.,  8.]]
        )
        >>> x[:, ::2] = -2.
        >>> x
        rn((2, 3), 'float64', 'pytorch').element(
            [[-2.,  7., -2.],
             [-2.,  7., -2.]]
        )
        """
        if isinstance(indices, type(self)):
            indices = indices.data
        if isinstance(values, type(self)):
            values = values.data

        if isinstance(values, (list, tuple)):
            values = self.array_backend.array_constructor(values)
        self.data[indices] = values

if __name__ == '__main__':
    from odl.core.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2019 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Unit tests for the ODL-PyTorch integration."""

import numpy as np
import torch
from torch import nn

import odl
from odl.contrib import torch as odl_torch
from odl.core.util.testutils import all_almost_equal, simple_fixture
from odl.core.util.dtype_utils import _universal_dtype_identifier

from odl.backends.arrays.pytorch_tensors import pytorch_array_backend 

dtype = simple_fixture('dtype', ['float32', 'float64'])
device_params = pytorch_array_backend.available_devices
device = simple_fixture('device', device_params)
shape = simple_fixture('shape', [(3,), (2, 3), (2, 2, 3)])


def test_autograd_function_forward(dtype, device, odl_impl_device_pairs):
    """Test forward evaluation with operators as autograd functions."""
    # Define ODL operator
    matrix = np.random.rand(2, 3)
    impl, odl_device = odl_impl_device_pairs
    space = odl.tensor_space((2,3), impl=impl, device=odl_device, dtype=dtype)
    matrix = space.element(matrix)
    odl_op = odl.MatrixOperator(matrix)

    # Compute forward pass with both ODL and PyTorch
    x_arr = np.ones(3, dtype=dtype)
    x_odl = odl_op.domain.element(x_arr)
    x = torch.from_numpy(x_arr).to(device)
    res = odl_torch.OperatorFunction.apply(odl_op, x)
    odl_res = odl_op(x_odl)
    odl_res_torch = torch.asarray(odl_res.data, device=device)
    assert _universal_dtype_identifier(res.dtype) == dtype
    assert all_almost_equal(res, odl_res_torch)
    assert str(x.device)== str(res.device) == device


def test_autograd_function_backward(dtype, device, odl_impl_device_pairs):
    """Test backprop with operators/functionals as autograd functions."""

    impl, odl_device = odl_impl_device_pairs

    # Define ODL operator and cost functional
    matrix = np.random.rand(2, 3).astype(dtype)
    odl_op = odl.MatrixOperator(matrix, impl=impl, device=odl_device)
    odl_cost = odl.functionals.L2NormSquared(odl_op.range)
    odl_functional = odl_cost * odl_op

    # Define evaluation point and mark as `requires_grad` to enable
    # backpropagation
    x_arr = np.ones(3, dtype=dtype)
    x = torch.from_numpy(x_arr).to(device)
    x.requires_grad_(True)

    # Compute forward pass
    y = odl_torch.OperatorFunction.apply(odl_op, x)
    res = odl_torch.OperatorFunction.apply(odl_cost, y)

    # Populate gradients by backwards pass
    res.backward()
    grad = x.grad
    grad_arr = grad.detach().cpu().numpy()

    # Compute gradient with ODL
    odl_grad = odl_functional.gradient(x_arr)

    assert grad_arr.dtype == dtype
    assert all_almost_equal(grad_arr, odl_grad)
    assert x.device == grad.device == torch.device(device)


def test_module_forward(shape, device, odl_impl_device_pairs):
    """Test forward evaluation with operators as modules."""

    impl, odl_device = odl_impl_device_pairs

    # Define ODL operator and wrap as module
    ndim = len(shape)
    space = odl.uniform_discr([0] * ndim, shape, shape, dtype='float32', impl=impl, device=odl_device)
    odl_op = odl.ScalingOperator(space, 2)
    op_mod = odl_torch.OperatorModule(odl_op)

    # Input data
    x_arr = np.ones(shape, dtype='float32')

    # Test with 1 extra dim (minimum)
    x = torch.from_numpy(x_arr).to(device)[None, ...]
    x.requires_grad_(True)
    res = op_mod(x)
    res_arr = res.detach().cpu().numpy()
    assert res_arr.shape == (1,) + odl_op.range.shape
    assert all_almost_equal(
        res_arr, odl_op(x_arr).data[None, ...]
    )
    assert x.device == res.device == torch.device(device)

    # Test with 2 extra dims
    x = torch.from_numpy(x_arr).to(device)[None, None, ...]
    x.requires_grad_(True)
    res = op_mod(x)
    res_arr = res.detach().cpu().numpy()
    assert res_arr.shape == (1, 1) + odl_op.range.shape
    assert all_almost_equal(
        res_arr, odl_op(x_arr).data[None, None, ...]
    )
    assert x.device == res.device == torch.device(device)


def test_module_forward_diff_shapes(device, odl_impl_device_pairs):
    """Test operator module with different shapes of input and output."""

    impl, odl_device = odl_impl_device_pairs

    # Define ODL operator and wrap as module
    matrix = np.random.rand(2, 3).astype('float32')
    odl_op = odl.MatrixOperator(matrix, impl=impl, device=odl_device)
    op_mod = odl_torch.OperatorModule(odl_op)

    # Input data
    x_arr = np.ones(3, dtype='float32')

    # Test with 1 extra dim (minimum)
    x = torch.from_numpy(x_arr).to(device)[None, ...]
    x.requires_grad_(True)
    res = op_mod(x)
    res_arr = res.detach().cpu().numpy()
    assert res_arr.shape == (1,) + odl_op.range.shape
    assert all_almost_equal(
        res_arr, odl_op(x_arr).data[None, ...]
    )
    assert x.device == res.device == torch.device(device)

    # Test with 2 extra dims
    x = torch.from_numpy(x_arr).to(device)[None, None, ...]
    x.requires_grad_(True)
    res = op_mod(x)
    res_arr = res.detach().cpu().numpy()
    assert res_arr.shape == (1, 1) + odl_op.range.shape
    assert all_almost_equal(
        res_arr, odl_op(x_arr).data[None, None, ...]
    )
    assert x.device == res.device == torch.device(device)


def test_module_backward(device, odl_impl_device_pairs):
    """Test backpropagation with operators as modules."""

    impl, odl_device = odl_impl_device_pairs

    # Define ODL operator and wrap as module
    matrix = np.random.rand(2, 3).astype('float32')
    odl_op = odl.MatrixOperator(matrix, impl=impl, device=odl_device)
    op_mod = odl_torch.OperatorModule(odl_op)
    loss_fn = nn.MSELoss()

    # Test with linear layers (1 extra dim)
    layer_before = nn.Linear(3, 3)
    layer_after = nn.Linear(2, 2)
    model = nn.Sequential(layer_before, op_mod, layer_after).to(device)
    x = torch.from_numpy(
        np.ones(3, dtype='float32')
    )[None, ...].to(device)
    x.requires_grad_(True)
    target = torch.from_numpy(
        np.zeros(2, dtype='float32')
    )[None, ...].to(device)
    loss = loss_fn(model(x), target)
    loss.backward()
    assert all(p is not None for p in model.parameters())
    assert x.grad.detach().cpu().abs().sum() != 0
    assert x.device == loss.device == torch.device(device)

    # Test with conv layers (2 extra dims)
    layer_before = nn.Conv1d(1, 2, 2)  # 1->2 channels
    layer_after = nn.Conv1d(2, 1, 2)  # 2->1 channels
    model = nn.Sequential(layer_before, op_mod, layer_after).to(device)
    # Input size 4 since initial convolution reduces by 1
    x = torch.from_numpy(
        np.ones(4, dtype='float32')
    )[None, None, ...].to(device)
    x.requires_grad_(True)
    # Output size 1 since final convolution reduces by 1
    target = torch.from_numpy(
        np.zeros(1, dtype='float32')
    )[None, None, ...].to(device)

    loss = loss_fn(model(x), target)
    loss.backward()
    assert all(p is not None for p in model.parameters())
    assert x.grad.detach().cpu().abs().sum() != 0
    assert x.device == loss.device == torch.device(device)


if __name__ == '__main__':
    odl.core.util.test_file(__file__)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Interface for sparse matrices"""

# We import to initialize the backends
# pylint: disable=unused-import
# We want to import if the backends are actually available
# pylint: disable=import-outside-toplevel
# We want to use a global statement here
# pylint: disable=global-statement

from typing import Optional
import importlib.util

from odl.core.sparse.sparse_template import (
    SparseMatrixFormat,
    _registered_sparse_formats,
)


IS_INITIALIZED = False

def _initialize_if_needed():
    """Initialize ``_registered_sparse_formats`` if not already done."""
    global IS_INITIALIZED
    if not IS_INITIALIZED:
        import odl.backends.sparse.scipy_backend

        torch_module = importlib.util.find_spec("torch")
        if torch_module is not None:
            try:
                import odl.backends.sparse.pytorch_backend
            except ModuleNotFoundError:
                pass
        IS_INITIALIZED = True


class SparseMatrix:
    """
    SparseMatrix is the ODL interface to the sparse Matrix supports in different backends.

    Note:
    The user is responsible for using the *args and **kwargs expected by the respective backends:
    Pytorch:
        -> COO: https://docs.pytorch.org/docs/stable/generated/torch.sparse_coo_tensor.html
    Scipy:
        -> COO: https://docs.scipy.org/doc/scipy/reference/generated/scipy.sparse.coo_matrix.

    Examples:
    SparseMatrix('COO', 'pytorch',
                [[0, 1, 1],[2, 0, 2]], [3, 4, 5],
                device='cuda:0')
    SparseMatrix('COO', 'scipy',
                (3, 4))
    """
    def __new__(cls,  format:str, impl:str, *args, **kwargs):

        _initialize_if_needed()

        # sanity checks
        assert isinstance(
            format, str
        ), f"The sparse data format can only be a string, got {type(format)}"
        assert isinstance(
            impl, str
        ), f"The impl argument can only be a str, got {type(impl)}"

        # Getting the backend (scipy, Pytorch...)
        backend_formats = _registered_sparse_formats.get(impl)
        if backend_formats is None:
            raise ValueError(
                f"The backend {impl} is not supported. Only {list(_registered_sparse_formats.keys())} are registered backends."
            )
        # Getting the format (COO, CSR...)
        sparse_impl = backend_formats.get(format)
        if sparse_impl is None:
            raise ValueError(
                f"No format {impl}. Only {list(backend_formats.keys())} are registered backends."
            )

        return sparse_impl.constructor(*args, **kwargs)


def lookup_sparse_format(matrix: object) -> Optional[SparseMatrixFormat]:
    """Looks up the sparse format of a matrix.
    Goes through the registered backends (scipy, pytorch...) and formats (COO, CSR...)

    Args:
        matrix (object): The matrix we want to get the sparse format of

    Returns:
        Optional[SparseMatrixFormat]: returns the sparse-format identifier if
        the matrix has one of the registered formats. Otherwise `None`.

    Notes:
        "sp_bkend" = sparse backend
        "sp_fmt"   = sparse format
    """
    _initialize_if_needed()
    for sp_bkend in _registered_sparse_formats.values():
        for sp_fmt in sp_bkend.values():
            if sp_fmt.is_of_this_sparse_format(matrix):
                return sp_fmt
    return None


def is_sparse(matrix:object) -> bool:
    """Checks whether the object is a sparse matrix in one
    of the format known to ODL.

    Args:
        matrix (object): input matrix

    Returns:
        bool: True if matrix is sparse else False
    """
    return lookup_sparse_format(matrix) is not None


def get_sparse_matrix_impl(matrix:object) -> str:
    """Gets the implementation string name of a matrix (which
    must be in one of the sparse formats known to ODL).

    Args:
        matrix (object): matrix

    Returns:
        str: The implementation string identifier ('pytorch', 'scipy', ...)
    """
    instance = lookup_sparse_format(matrix)
    assert instance is not None, "The matrix is not a supported sparse matrix"
    return instance.impl


def get_sparse_matrix_format(matrix:object) -> str:
    """Gets the format string name of a matrix

    Args:
        matrix (object): matrix

    Returns:
        str: The format string identifier ('COO', 'CSR', ...)
    """
    instance = lookup_sparse_format(matrix)
    assert instance is not None, "The matrix is not a supported sparse matrix"
    return instance.sparse_format

if __name__ == '__main__':
    print(SparseMatrix('COO', 'scipy', (3, 4)))



"""ODL wrapper for multi-backend Sparse Matrices support."""

from dataclasses import dataclass
from typing import Callable

_registered_sparse_formats = {}


@dataclass
class SparseMatrixFormat:
    """Metainformation about Sparse Matrices in some storage format.
    This is a singleton class, it has a post-init method that registers the Sparse Matrix
    format in a global, internal dict
    (with [impl][format] key structure, for instance ['scipy']['COO'])

    Attributes
    ----------
    sparse_format : str
        The shorthand descriptor of the layout of the sparse matrix, e.g 'COO'
    impl : str
        The identifier of the backend, e.g 'scipy'
    constructor : Callable
        Constructor of the Sparse backend
    is_of_this_sparse_format : Callable[[object], bool]
        Checks whether an array is of the same backend and layout of the given SparseMatrixFormat 
    to_dense : Callable
        Makes a sparse matrix dense
    matmul_spmatrix_with_vector : Callable
        Sparse matrix multiplication function (the signature differs accross backends)
    """
    sparse_format: str
    impl: str
    constructor: Callable
    is_of_this_sparse_format: Callable[[object], bool]
    to_dense: Callable
    matmul_spmatrix_with_vector: Callable

    def __post_init__(self):
        if self.impl not in _registered_sparse_formats:
            _registered_sparse_formats[self.impl] = {}
        if self.sparse_format in _registered_sparse_formats[self.impl]:
            raise KeyError(
                f"A {self.sparse_format} sparse format for backend {self.impl}"
               + " is already registered."
               + " Every sparse format needs to have a unique identifier combination."
            )
        _registered_sparse_formats[self.impl][self.sparse_format] = self



"""Source code for the getting started example."""

import odl
import scipy.signal


class Convolution(odl.Operator):
    """Operator calculating the convolution of a kernel with a function.

    The operator inherits from ``odl.Operator`` to be able to be used with ODL.
    """

    def __init__(self, kernel):
        """Initialize a convolution operator with a known kernel."""

        # Store the kernel
        self.kernel = kernel

        # Initialize the Operator class by calling its __init__ method.
        # This sets properties such as domain and range and allows the other
        # operator convenience functions to work.
        super(Convolution, self).__init__(
            domain=kernel.space, range=kernel.space, linear=True)

    def _call(self, x):
        """Implement calling the operator by calling scipy."""
        return scipy.signal.fftconvolve(self.kernel, x, mode='same')

    @property  # making adjoint a property lets users access it as A.adjoint
    def adjoint(self):
        return self  # the adjoint is the same as this operator


# Define the space the problem should be solved on.
# Here the square [-1, 1] x [-1, 1] discretized on a 100x100 grid.
space = odl.uniform_discr([-1, -1], [1, 1], [100, 100])

# Convolution kernel, a small centered rectangle.
kernel = odl.core.phantom.cuboid(space, [-0.05, -0.05], [0.05, 0.05])

# Create convolution operator
A = Convolution(kernel)

# Create phantom (the "unknown" solution)
phantom = odl.core.phantom.shepp_logan(space, modified=True)

# Apply convolution to phantom to create data
g = A(phantom)

# Display the results using the show method
kernel.show('kernel')
phantom.show('phantom')
g.show('convolved phantom')

# Landweber

# Need operator norm for step length (omega)
opnorm = odl.power_method_opnorm(A)

f = space.zero()
odl.solvers.landweber(A, f, g, niter=100, omega=1 / opnorm ** 2)
f.show('landweber')

# Conjugate gradient

f = space.zero()
odl.solvers.conjugate_gradient_normal(A, f, g, niter=100)
f.show('conjugate gradient')

# Tikhonov with identity

B = odl.IdentityOperator(space)
a = 0.1
T = A.adjoint * A + a * B.adjoint * B
b = A.adjoint(g)

f = space.zero()
odl.solvers.conjugate_gradient(T, f, b, niter=100)
f.show('Tikhonov identity conjugate gradient')

# Tikhonov with gradient

B = odl.Gradient(space)
a = 0.0001
T = A.adjoint * A + a * B.adjoint * B
b = A.adjoint(g)

f = space.zero()
odl.solvers.conjugate_gradient(T, f, b, niter=100)
f.show('Tikhonov gradient conjugate gradient')

# Douglas-Rachford

# Assemble all operators into a list.
grad = odl.Gradient(space)
lin_ops = [A, grad]
a = 0.001

# Create functionals for the l2 distance and l1 norm.
g_funcs = [odl.functionals.L2NormSquared(space).translated(g),
           a * odl.functionals.L1Norm(grad.range)]

# Functional of the bound constraint 0 <= f <= 1
f = odl.functionals.IndicatorBox(space, 0, 1)

# Find scaling constants so that the solver converges.
# See the douglas_rachford_pd documentation for more information.
opnorm_A = odl.power_method_opnorm(A, xstart=g)
opnorm_grad = odl.power_method_opnorm(grad, xstart=g)
sigma = [1 / opnorm_A**2, 1 / opnorm_grad**2]
tau = 1.0

# Solve using the Douglas-Rachford Primal-Dual method
x = space.zero()
odl.solvers.douglas_rachford_pd(x, f, g_funcs, lin_ops,
                                tau=tau, sigma=sigma, niter=100)
x.show('TV Douglas-Rachford', force_show=True)



# Copyright 2014-2025 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

# pylint: disable=line-too-long
# pylint: disable=invalid-name

"""PyTorch implementation of Sparse Matrices."""

from torch import sparse_coo_tensor, Tensor, sparse_coo, matmul

from odl.core.sparse.sparse_template import SparseMatrixFormat, _registered_sparse_formats

def is_sparse_COO(matrix):
    """Convenience function to check whether a matrix is a sparse PyTorch Tensor with the COO format"""
    return isinstance(matrix, Tensor) and matrix.is_sparse and matrix.layout == sparse_coo

if ('pytorch' not in _registered_sparse_formats
        or 'COO' not in _registered_sparse_formats['pytorch']):
    pytorch_coo_tensor = SparseMatrixFormat(
        sparse_format='COO',
        impl = 'pytorch',
        constructor = sparse_coo_tensor,
        is_of_this_sparse_format = is_sparse_COO,
        to_dense = lambda matrix: matrix.to_dense(),
        matmul_spmatrix_with_vector = matmul
    )



# Copyright 2014-2017 The ODL contributors
#
# This file is part of ODL.
#
# This Source Code Form is subject to the terms of the Mozilla Public License,
# v. 2.0. If a copy of the MPL was not distributed with this file, You can
# obtain one at https://mozilla.org/MPL/2.0/.

"""Tests for unsupervised FoMs."""

import numpy as np
import pytest
import odl
import odl.contrib.fom


def noise_element(space):
    return space.element(np.random.randn(*space.shape))

def test_estimate_noise_std_constant_1d():
    """Verify ``estimate_noise_std(0) == 0``"""
    img = odl.rn(10).zero()
    result = odl.contrib.fom.estimate_noise_std(img)
    assert pytest.approx(result) == 0.0


def test_estimate_noise_std_normal_1d():
    """Verify ``estimate_noise_std(N(0, 1)) == 1`` in 1d."""
    img = noise_element(odl.rn(1000))
    result = odl.contrib.fom.estimate_noise_std(img)
    expected = np.std(img.asarray())
    assert pytest.approx(result, abs=0.2) == expected


def test_estimate_noise_std_normal_2d():
    """Verify ``estimate_noise_std(N(0, 1)) == 1`` in 2d."""
    img = noise_element(odl.rn((100, 100)))
    result = odl.contrib.fom.estimate_noise_std(img)
    expected = np.std(img.asarray())
    assert pytest.approx(result, abs=0.2) == expected


def test_estimate_noise_std_normal_4d():
    """Verify ``estimate_noise_std(N(0, 1)) == 1`` in 4d."""
    img = noise_element(odl.rn((10, 10, 10, 10)))
    result = odl.contrib.fom.estimate_noise_std(img)
    expected = np.std(img.asarray())
    assert pytest.approx(result, abs=0.2) == expected


def test_estimate_noise_std_normal_large_1d():
    """Verify ``estimate_noise_std(N(0, 1)) == 1`` with low error."""
    img = noise_element(odl.rn(100000))
    result = odl.contrib.fom.estimate_noise_std(img)
    expected = np.std(img.asarray())
    assert pytest.approx(result, abs=0.01) == expected


def test_estimate_noise_std_normal_2d_pointwise():
    """Verify ``estimate_noise_std(N(0, 1)) == 1`` in 2d."""
    img = noise_element(odl.rn((100, 100)))
    result = odl.contrib.fom.estimate_noise_std(img, average=False)
    result_mean = np.mean(result.asarray())
    expected = np.std(img.asarray())
    assert result.shape == img.shape
    assert result.dtype == result.dtype
    assert pytest.approx(result_mean, abs=0.25) == expected


if __name__ == '__main__':
    odl.core.util.test_file(__file__)
